import re
from cachetools import Cache as _Cache
import inspect
from typing import Any
from loguru import logger


class _BaseCache:

    def __init__(self, cache: _Cache, debug=False):
        """
        Args:
            cache: 缓存管理对象
            debug: 调试模式, 调试模式会设置日志
        """
        self._cache = cache
        self._none = object()
        self._debug = debug

    def _wrapper_data(self, data):
        if data is None:
            return self._none
        return data

    def _unwrapper_data(self, data):
        if data is self._none:
            return None
        return data

    def _get_key_name_value(self, obj, key: str):
        keys = key.split('.')
        for k in keys:
            if isinstance(obj, dict):
                obj = obj.get(k, self._none)
            else:
                try:
                    obj = getattr(obj, k)
                except AttributeError:
                    obj = self._none
            if obj is self._none:
                raise ValueError('缓存key[{}]中的属性[{}]不存在'.format(key, k))
        return obj

    def _gen_key(self, func, key, *args, **kwargs) -> str:
        """
        将key中的变量值进行替换
        """
        mapping = {k: v for k, v in kwargs.items()}
        # args映射
        if args:
            signature = inspect.signature(func)
            names = list(signature.parameters.keys())
            for i in range(min(len(names), len(args))):
                mapping[names[i]] = args[i]
        # 定义正则表达式匹配${}包含的内容
        pattern = r'\${(.*?)}'
        # 使用正则表达式查找所有匹配的内容
        matches = re.findall(pattern, key)
        # 遍历匹配结果
        for match in matches:
            # 如果匹配到的内容在字典中存在对应的映射，则进行替换
            key = key.replace(f'${{{match}}}', str(self._get_key_name_value(mapping, match)))
        return key

    def put(self, key: str, data: Any):
        """设置缓存

        Args:
            key: 缓存的key
            data: 缓存的数据
        """
        self._debug_msg('设置缓存key[{}], 值: {}', key, data)
        self._cache[key] = self._wrapper_data(data)

    def get(self, key: str, default=None) -> Any:
        """获取缓存值

        Args:
            key: 缓存的key值
            default: 若缓存不存在数据, 则返回的默认值

        Returns:
            若该缓存存在, 则返回该返回的值, 否则返回default默认值
        """
        data = self._unwrapper_data(self._cache.get(key, default=default))
        self._debug_msg('获取缓存key[{}], 值: {}', key, data)
        return data

    def evict(self, key: str, default=None) -> Any:
        """清除缓存

        Args:
            key: 缓存的key
            default: 若缓存不存在数据, 则返回的默认值
        Returns:
            若该缓存存在, 则返回该返回的值, 否则返回default默认值
        """
        data = self._unwrapper_data(self._cache.pop(key, default=default))
        self._debug_msg('删除缓存key[{}], 值: {}', key, data)
        return data

    def clear(self):
        """清空所有缓存"""
        self._cache.clear()
        self._debug_msg('清空所有缓存')

    def _debug_msg(self, msg: str, *args, **kwargs):
        if self._debug:
            logger.debug(msg, *args, **kwargs)


class Cache(_BaseCache):

    def __init__(self, cache: _Cache, debug=False):
        super().__init__(cache, debug)

    def cacheable(self, key: str):
        """缓存查询装饰器, 参数变量使用${参数名}填充
        使用示例:
            from cachetools import TTLCache

            cache = Cache(cache=TTLCache(maxsize=1024, ttl=600))

            @cache.cacheable(key='xxx_${a}_${b}')
            def xxx(a: int, b: str):
                # 业务逻辑

            # 调用
            v = xxx(1, 's') # 第一次调用将结果设置到key为 xxx_1_s 的缓存上, 再次调用将直接返回缓存结果

        Args:
            key: key的表达式, 使用函数的变量使用${}包裹
        """

        def wrapper_func(func):
            def wrapper(*args, **kwargs):
                _key = self._gen_key(func, key, *args, **kwargs)
                data = self._cache.get(_key)
                if data is not None:
                    data = self._unwrapper_data(data)
                    self._debug_msg('获取缓存key[{}], 值: {}', _key, data)
                    return data
                data = func(*args, **kwargs)
                self.put(_key, data)
                return data

            return wrapper

        return wrapper_func

    def cache_put(self, key: str):
        """缓存设置装饰器, 参数变量使用${参数名}填充
        使用示例:
            from cachetools import TTLCache

            cache = Cache(cache=TTLCache(maxsize=1024, ttl=600))

            @cache.cache_put(key='xxx_${a}_${b}')
            def xxx(a: int, b: str):
                # 业务逻辑

            # 调用
            xxx(1, 's') # 每次调用都将方法返回的结果 设置到缓存key为 xxx_1_s 上
        """

        def wrapper_func(func):
            def wrapper(*args, **kwargs):
                _key = self._gen_key(func, key, *args, **kwargs)
                data = func(*args, **kwargs)
                self.put(_key, data)
                return data

            return wrapper

        return wrapper_func

    def cache_evict(self, key: str):
        """缓存清除装饰器, 参数变量使用${参数名}填充
        使用示例:
            from cachetools import TTLCache

            cache = Cache(cache=TTLCache(maxsize=1024, ttl=600))

            @cache.cache_evict(key='xxx_${a}_${b}')
            def xxx(a: int, b: str):
                # 业务逻辑

            # 调用
            xxx(1, 's') # 清除key为 xxx_1_s 的缓存

        Args:
            key: key的表达式, 使用函数的变量使用${}包裹
        """

        def wrapper_func(func):
            def wrapper(*args, **kwargs):
                _key = self._gen_key(func, key, *args, **kwargs)
                self.evict(_key, default=None)
                return func(*args, **kwargs)

            return wrapper

        return wrapper_func


class AsyncCache(_BaseCache):

    def __init__(self, cache: _Cache, debug=False):
        super().__init__(cache, debug)

    def cacheable(self, key: str):
        """缓存查询装饰器, 参数变量使用${参数名}填充
        使用示例:
            from cachetools import TTLCache

            async_cache = AsyncCache(cache=TTLCache(maxsize=1024, ttl=600))

            @async_cache.cacheable(key='xxx_${a}_${b}')
            async def xxx(a: int, b: str):
                # 业务逻辑

            # 调用
            v = xxx(1, 's') # 第一次调用将结果设置到key为 xxx_1_s 的缓存上, 再次调用将直接返回缓存结果

        Args:
            key: key的表达式, 使用函数的变量使用${}包裹
        """

        def wrapper_func(async_func):
            async def wrapper(*args, **kwargs):
                _key = self._gen_key(async_func, key, *args, **kwargs)
                data = self._cache.get(_key)
                if data is not None:
                    data = self._unwrapper_data(data)
                    self._debug_msg('获取缓存key[{}], 值: {}', _key, data)
                    return data
                data = await async_func(*args, **kwargs)
                self.put(_key, data)
                return data

            return wrapper

        return wrapper_func

    def cache_put(self, key: str):
        """缓存设置装饰器, 参数变量使用${参数名}填充
        使用示例:
            from cachetools import TTLCache

            async_cache = AsyncCache(cache=TTLCache(maxsize=1024, ttl=600))

            @async_cache.cache_put(key='xxx_${a}_${b}')
            def xxx(a: int, b: str):
                # 业务逻辑

            # 调用
            xxx(1, 's') # 每次调用都将方法返回的结果 设置到缓存key为 xxx_1_s 上
        """

        def wrapper_func(async_func):
            async def wrapper(*args, **kwargs):
                _key = self._gen_key(async_func, key, *args, **kwargs)
                data = await async_func(*args, **kwargs)
                self.put(_key, data)
                return data

            return wrapper

        return wrapper_func

    def cache_evict(self, key: str):
        """缓存清除装饰器, 参数变量使用${参数名}填充
        使用示例:
            from cachetools import TTLCache

            async_cache = AsyncCache(cache=TTLCache(maxsize=1024, ttl=600))

            @async_cache.cache_evict(key='xxx_${a}_${b}')
            def xxx(a: int, b: str):
                # 业务逻辑

            # 调用
            xxx(1, 's') # 清除key为 xxx_1_s 的缓存

        Args:
            key: key的表达式, 使用函数的变量使用${}包裹
        """

        def wrapper_func(async_func):
            async def wrapper(*args, **kwargs):
                _key = self._gen_key(async_func, key, *args, **kwargs)
                self.evict(_key, default=None)
                return await async_func(*args, **kwargs)

            return wrapper

        return wrapper_func



import hashlib
from typing import Union
from enum import Enum
from .base import _convert_data


def _md5_16(data: bytes):
    # 计算32位MD5哈希值
    return hashlib.md5(data).hexdigest()[8:-8]


def _md5_32(data: bytes):
    # 计算32位MD5哈希值
    return hashlib.md5(data).hexdigest()


def _md5_64(data: bytes):
    # 计算32位MD5哈希值
    return hashlib.md5(data).hexdigest() * 2


class Md5DigitEnum(Enum):
    MD5_16 = 16
    MD5_32 = 32
    MD5_64 = 64


_DIGIT_HANDLE_MAP = {
    Md5DigitEnum.MD5_16: _md5_16,
    Md5DigitEnum.MD5_32: _md5_32,
    Md5DigitEnum.MD5_64: _md5_64,
}


def md5(data: Union[str, bytes],
        encoding='utf-8',
        digit: Md5DigitEnum = Md5DigitEnum.MD5_32) -> str:
    """MD5 加密工具

    Args:
        data: 需要加密的数据
        encoding: 数据的字符集类型
        digit: 加密的位数
    """
    data = _convert_data(data, encoding)
    handle = _DIGIT_HANDLE_MAP.get(digit)
    if not handle:
        raise ValueError('不支持的md5加密位数: {}'.format(digit))
    return handle(data)



from enum import Enum
from typing import Union, Type, Optional

from pydantic import BaseModel

from .base import DataFileLoader
from .csv import CsvDataFileLoader
from .yaml import YamlDataFileLoader
from .json import JsonDataFileLoader
from .ini import IniDataFileLoader
from .eval import EvalDataFileLoader
from .properties import PropertiesDataFileLoader

_csv_data_file_loader = CsvDataFileLoader()
_yaml_data_file_loader = YamlDataFileLoader()
_json_data_file_loader = JsonDataFileLoader()
_ini_data_file_loader = IniDataFileLoader()
_properties_data_file_loader = PropertiesDataFileLoader()
_eval_data_file_loader = EvalDataFileLoader()


class DataType(Enum):
    """ 数据类型 """
    JSON = (('json',), _json_data_file_loader.load,
            _json_data_file_loader.load_file)
    YAML = (('yaml', 'yml'), _yaml_data_file_loader.load,
            _yaml_data_file_loader.load_file)
    INI = (('ini', 'toml', 'cfg'), _ini_data_file_loader.load,
           _ini_data_file_loader.load_file)
    PROPERTIES = (('properties',), _properties_data_file_loader.load,
                  _properties_data_file_loader.load_file)
    CSV = (('csv', ), _csv_data_file_loader.load,
           _csv_data_file_loader.load_file)
    EVAL = (('eval',), _eval_data_file_loader.load,
            _eval_data_file_loader.load_file)


def get_file_data_type(file_path: str) -> Optional[DataType]:
    """获取文件的数据类型枚举, 若识别不了则返回None

    Args:
        file_path: 文件路径
    """
    file_type = file_path.split('.')[-1]
    file_data_type = None
    for data_type in DataType:
        if file_type in data_type.value[0]:
            file_data_type = data_type
            break
    return file_data_type


class AutoDataFileLoader(DataFileLoader):
    """
    自动检测数据/文件类型加载器
    """

    def load(self,
             config_data: Union[str, bytes],
             modelclass: Union[Type[dict], Type[BaseModel]] = dict,
             data_type: DataType = None,
             header=False,
             **kwargs) -> Union[dict, list, BaseModel]:
        """加载配置数据到python或pydantic模型
        Args:
            config_data: 数据字符串
            modelclass: 模型类型
            data_type: 数据类型, None 表示自动解析 (若知道数据类型推荐传入指定数据类型, 防止自动推断类型错误)
            header: 第一行是否为表头, 仅支持CSV传参
        """
        if data_type:
            return data_type.value[1](config_data,
                                      modelclass=modelclass,
                                      header=header,
                                      use_header_as_key=header)

        for data_type in DataType:
            try:
                return data_type.value[1](config_data,
                                          modelclass=modelclass,
                                          header=header,
                                          use_header_as_key=header)
            except Exception:
                pass
        raise ValueError('无法解析的数据: ' + config_data)

    def load_file(self,
                  file_path: str,
                  encoding: str = 'utf-8',
                  modelclass: Union[Type[dict], Type[BaseModel]] = dict,
                  **kwargs) -> Union[dict, list, BaseModel]:
        """加载文件数据到python对象或pydantic模型

        Args:
            file_path: 文件路径
            encoding: 文件编码类型
            modelclass: 模型类型
        """
        data_type = get_file_data_type(file_path)
        if data_type:
            return data_type.value[2](file_path,
                                      encoding=encoding,
                                      modelclass=modelclass,
                                      **kwargs)

        for data_type in DataType:
            try:
                return data_type.value[2](file_path,
                                          encoding=encoding,
                                          modelclass=modelclass,
                                          **kwargs)
            except Exception:
                pass
        raise ValueError('无法解析的文件: ' + file_path)



from typing import Type, Dict, Any
from dragons96_tools.ioc.config import cfg
from dragons96_tools.ioc.utils.value_utils import convert


class Environment:

    def get_property(self, name: str, cls: Type = None) -> Any:
        if not name:
            raise ValueError('Environment name不能为空')
        data = self._parse_config_value(data=cfg(), config_name_path=name)
        if data is None:
            return None
        if cls:
            return convert(data, cls)
        return data

    def _parse_config_value(self, data: Dict, config_name_path: str):
        config_names = config_name_path.split('.')
        current = dict(data)
        for i, config_name in enumerate(config_names):
            if not isinstance(current, dict):
                return None
            if '[' in config_name:
                index_key, index_value = config_name.split('[')
                index_value = index_value.rstrip(']')
                current = current.get(index_key)
                if current is None:
                    return None
                if not isinstance(current, list):
                    return None
                if index_value >= len(current):
                    return None
                current = current[index_value]
            else:
                current = current.get(config_name)
        return current



from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad


def encrypt_aes(data, key, *args, **kwargs):
    """AES 加密

    Args:
        data: 数据
        key: 加密key

    Returns:
        加密后的bytes与iv值
    """
    cipher = AES.new(key, AES.MODE_CBC, *args, **kwargs)
    ciphertext = cipher.encrypt(pad(data, AES.block_size))
    iv = cipher.iv
    return ciphertext, iv


def decrypt_aes(ciphertext, key, iv, *args, **kwargs) -> bytes:
    """AES 解密

    Args:
        ciphertext: 解密串bytes
        iv: iv值
        key: 加密的key

    Returns:
        解密后的数据
    """
    cipher = AES.new(key, AES.MODE_CBC, iv=iv, *args, **kwargs)
    plaintext = unpad(cipher.decrypt(ciphertext), AES.block_size)
    return plaintext



from pydantic import BaseModel as BM, ConfigDict, Field
from abc import ABC
import datetime
from typing import Optional, Any, List


class BaseModel(BM, ABC):
    """符合业务场景的通用基础Model"""

    model_config = ConfigDict(json_encoders={
        datetime.datetime: lambda dt: dt.strftime('%Y-%m-%d %H:%M:%S')
    })

    def __init__(self, **kwargs):
        # 针对str类型, 额外支持int, float, bool类型转换
        for field_name, field_type in self.__annotations__.items():
            if field_type == str or field_type == Optional[str]:
                if field_name in kwargs:
                    value = kwargs[field_name]
                    if isinstance(value, (int, float)):
                        kwargs[field_name] = str(value)
                    elif isinstance(value, bool):
                        kwargs[field_name] = str(value).lower()
        super().__init__(**kwargs)


class R(BaseModel):
    """通用响应对象"""
    # 状态码
    code: int = Field(..., title='状态码')
    # 状态信息
    msg: str = Field(..., title='状态信息')
    # 响应数据
    data: Optional[Any] = Field(None, title='响应数据')

    @staticmethod
    def ok(data: Optional[Any] = None, msg: str = '请求成功', code: int = 200):
        return R(code=code, msg=msg, data=data)

    @staticmethod
    def fail(msg: str = '请求失败', code: int = 500):
        return R(code=code, msg=msg, data=None)


class PageModel(BaseModel):
    """分页Model, 可作分页入参可作分页出参, 入参仅需传入page, 与page_size, 出参则均需传入"""
    # 分页数据记录列表
    rows: Optional[List[Any]] = Field([], title='分页数据记录列表')
    # 当前页码
    page: Optional[int] = Field(..., title='当前页码')
    # 当前页大写
    page_size: Optional[int] = Field(..., title='当前页大小')
    # 记录总数
    total: Optional[int] = Field(..., title='记录总数')


class PageR(R):
    """通用分页响应"""
    data: Optional[PageModel] = Field(..., title='响应数据')

    @staticmethod
    def ok(data: Optional[PageModel] = None, msg: str = '请求成功', code: int = 200):
        if not data:
            data = PageModel()
        return R.ok(code=code, msg=msg, data=data)



from typing import Any
from .base import BaseBeanProxy
from dragons96_tools.ioc.constants import required


class DelayAutowiredClassBeanProxy(BaseBeanProxy):
    """延迟类bean代理"""

    def __init__(self, name: str, obj: Any, *args, **kwargs):
        """
        Args:
            obj: class 类型
        """
        super().__init__(name, None, *args, **kwargs)
        self._cls = obj

    def _delay_inject_obj(self):
        from dragons96_tools.ioc import get_application_context
        obj = get_application_context().get_bean(name=self._name, cls=self._cls)
        if self._extra_kwargs.get('required', True):
            if not obj:
                raise ValueError(
                    f'无法注入{("名称[" + self._name + "]") if self._name else ""}{("类型[" + self._cls + "]") if self._cls else ""}的bean')
        return obj

    def get(self):
        if self._obj is None:
            self._obj = self._delay_inject_obj()
        return self._obj

    def ioc_type(self):
        return self._cls


class DelayConfigAutowiredClassBeanProxy(BaseBeanProxy):

    def __init__(self, name: str, obj: Any, *args, **kwargs):
        """
        Args:
            name: 配置名称
            obj: 类型
        """
        super().__init__(name, None, *args, **kwargs)
        self._cls = obj

    def _delay_inject_obj(self):
        from dragons96_tools.ioc import get_environment
        env = get_environment()
        # 非ioc环境直接返回
        if not env:
            return None
        if not (self._name.startswith('${') and self._name.endswith('}')):
            raise ValueError('配置必须使用${}包裹')
        name = self._name[2:-1]
        v = env.get_property(name, cls=self._cls)
        if v is None:
            if self._extra_kwargs.get('default_value', required) == required:
                raise ValueError(f'{self._name}配置属性获取失败, 请检查配置')
            return self._extra_kwargs.get('default_value')
        return v

    def get(self):
        if self._obj is None:
            self._obj = self._delay_inject_obj()
        return self._obj

    def ioc_type(self):
        return self._cls



import time
from loguru import logger

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class _BaseRetry:
    """重试基类"""

    def __init__(self, max_attempts=3, delay=1,
                 log_enable: bool = True,
                 ever_fail_level: Literal['TRACE', 'DEBUG', 'INFO', 'WARNING', 'ERROR'] = 'WARNING',
                 fail_level: Literal['TRACE', 'DEBUG', 'INFO', 'WARNING', 'ERROR'] = 'ERROR'):
        """初始化重试工具

        Args:
            max_attempts (int): 最大重试次数，默认为 3 次。
            delay (float): 重试之间的延迟时间，单位为秒，默认为 1 秒。
            log_enable: 是否记录日志
            ever_fail_level: 每次失败记录的日志级别
            fail_level: 全部失败记录的日志级别
        """
        self._max_attempts = max_attempts
        self._delay = delay
        self._log_enable = log_enable
        self._ever_fail_level = ever_fail_level
        self._fail_level = fail_level

    def _handle_attempt_error(self, attempt, func, e, *args, **kwargs):
        """处理一次重试失败

        Args:
            attempt: 第attempt次失败, 从1开始
            func: 执行的函数
            e: 发生的异常
            args: 执行的函数的args参数
            kwargs: 执行的函数的kwargs参数
        """
        if attempt < self._max_attempts:
            if self._log_enable:
                logger.log(self._ever_fail_level,
                           "函数[{}]第[{}]次执行失败, 失败原因: {}, 等待[{}]秒后将进行重试, 函数参数args: {}, kwargs: {}",
                           func.__name__, attempt, e, self._delay, args, kwargs)
            time.sleep(self._delay)
        else:
            if self._log_enable:
                logger.log(self._fail_level,
                           '函数[{}]重试[{}]次仍然失败, 抛出异常: {}, 函数参数args: {}, kwargs: {}',
                           func.__name__, self._max_attempts, e, args, kwargs)
            raise e


class Retry(_BaseRetry):
    """重试工具"""

    def __init__(self, max_attempts=3, delay=1, log_enable: bool = True,
                 ever_fail_level: Literal['TRACE', 'DEBUG', 'INFO', 'WARNING', 'ERROR'] = 'WARNING',
                 fail_level: Literal['TRACE', 'DEBUG', 'INFO', 'WARNING', 'ERROR'] = 'ERROR'):
        super().__init__(max_attempts, delay, log_enable, ever_fail_level, fail_level)

    def retry(self, func):
        """重试装饰器
        使用方式:
            retry = Retry(...)

            @retry.retry
            def xxx(*args, **kwargs):
                # 业务逻辑
        Args:
            func: 需要重试的函数
        """
        def wrapper(*args, **kwargs):
            return self.do(func, *args, **kwargs)
        return wrapper

    def do(self, func, *args, **kwargs):
        """重试函数，重复执行指定的函数，直到达到最大重试次数或者函数执行成功为止。

        Args:
            func: 要执行的函数。
            args: 执行函数的参数
            kwargs: 执行函数的参数

        Returns:
            函数执行成功时返回函数的返回值，否则返回 None。
        """
        for attempt in range(1, self._max_attempts + 1):
            try:
                result = func(*args, **kwargs)
                return result  # 函数执行成功，直接返回结果
            except Exception as e:
                super()._handle_attempt_error(attempt, func, e, *args, **kwargs)


class AsyncRetry(_BaseRetry):
    """重试工具Async"""

    def __init__(self, max_attempts=3, delay=1, log_enable: bool = True,
                 ever_fail_level: Literal['TRACE', 'DEBUG', 'INFO', 'WARNING', 'ERROR'] = 'WARNING',
                 fail_level: Literal['TRACE', 'DEBUG', 'INFO', 'WARNING', 'ERROR'] = 'ERROR'):
        super().__init__(max_attempts, delay, log_enable, ever_fail_level, fail_level)

    def retry(self, async_func):
        """重试装饰器
        使用方式:
            retry = AsyncRetry(...)

            @retry.retry
            async def xxx(*args, **kwargs):
                # 业务逻辑
        Args:
            async_func: 需要重试的async函数
        """
        async def wrapper(*args, **kwargs):
            return await self.do(async_func, *args, **kwargs)
        return wrapper

    async def do(self, async_func, *args, **kwargs):
        """重试函数，重复执行指定的函数，直到达到最大重试次数或者函数执行成功为止。

        Args:
            async_func: 要执行的async函数。
            args: 执行函数的参数
            kwargs: 执行函数的参数

        Returns:
            函数执行成功时返回函数的返回值，否则返回 None。
        """
        for attempt in range(1, self._max_attempts + 1):
            try:
                result = await async_func(*args, **kwargs)
                return result  # 函数执行成功，直接返回结果
            except Exception as e:
                super()._handle_attempt_error(attempt, async_func, e, *args, **kwargs)
